package com.syjy.calculate.conotroller;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.collection.ListUtil;
import cn.hutool.core.map.MapUtil;
import cn.hutool.core.util.NumberUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.ReUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.json.JSONObject;
import cn.hutool.json.JSONUtil;
import com.syjy.calculate.entity.*;
import com.syjy.calculate.repository.CalculationFormulaRepository;
import com.syjy.calculate.service.AccuracyPassRateCalculateService;
import com.syjy.calculate.service.AssessmentCalculateService;
import com.syjy.calculate.service.ComputeService;
import com.syjy.calculate.util.Common;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import javax.servlet.http.HttpServletRequest;
import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 公式管理
 *
 * @author zcl
 * @version 1.0
 * @since 2022/8/8 9:30
 */
@RestController
@RequestMapping("apiCalculate")
public class ApiCalculate {
  @Autowired
  CalculationFormulaRepository calculationFormulaRepository;
  @Autowired
  private AccuracyPassRateCalculateService accuracyPassRateCalculateService;
  @Autowired
  private AssessmentCalculateService assessmentCalculateService;
  @Autowired
  private Common common;
  @Autowired
  private ComputeService computeService;

  /**
   * 计算单个公式准确率
   *
   * @param request
   * @return
   */
  @PostMapping("/calculate")
  public CalculateResult calculate(HttpServletRequest request, @RequestBody JSONObject data) {
    // 校验信息
    CalculateResult checkResult = checkAuthorization(request);
    // 如果校验不成功，返回失败信息
    if (!CalculateResult.Type.SUCCESS.value().equals(String.valueOf(checkResult.get(CalculateResult.CODE_TAG)))) {
      return checkResult;
    }
    CalculateRequest calculateRequest;
    // 将传入的数据转为bean
    calculateRequest = data.toBean(CalculateRequest.class);
    // 调用计算方法
    return accuracyPassRateCalculateService.calculate(calculateRequest);
  }

  @PostMapping("/assessment")
  public CalculateResult assessment(HttpServletRequest request, @RequestBody JSONObject data) {
    // 校验信息
    CalculateResult checkResult = checkAuthorization(request);
    // 如果校验不成功，返回失败信息
    if (!CalculateResult.Type.SUCCESS.value().equals(String.valueOf(checkResult.get(CalculateResult.CODE_TAG)))) {
      return checkResult;
    }
    AssessmentRequest calculateRequest;
    // 将传入的数据转为bean
    calculateRequest = data.toBean(AssessmentRequest.class);
    // 调用计算方法
    CalculateResult resultMap = assessmentCalculateService.calculateAssessment(calculateRequest);
    return resultMap;
  }


  @PostMapping("/province")
  @SuppressWarnings({"rawtypes", "unchecked"})
  public List<Map<String, Object>> calculateByProvince(HttpServletRequest request, @RequestBody JSONObject data) {
    List<Map<String, Object>> result = new ArrayList<>();
    // 校验信息
    CalculateResult checkResult = checkAuthorization(request);
    // 如果校验不成功，返回失败信息
    if (!CalculateResult.Type.SUCCESS.value().equals(String.valueOf(checkResult.get(CalculateResult.CODE_TAG)))) {
      Map<String, Object> map = MapUtil.builder(new HashMap<String, Object>()).put("error", checkResult).build();
      result.add(map);
      return result;
    }
    //按省份获取公式
    List<CalculationFormula> formulaList = calculationFormulaRepository.findFormulaWithProvince(data.get("province", String.class), data.get("electricType", String.class));
    List<CalculationFormula> formulaAsms = new ArrayList<>();// 准确率公式
    List<CalculationFormula> formulaAccs = new ArrayList<>();// 考核公式
    //
    List<CalculationFormula> formulas = formulaList.stream().filter(t -> {// 过滤，排除版本，规则和用户
      boolean v = !StrUtil.equalsIgnoreCase("VERSION", t.getType());
      boolean r = !StrUtil.equalsIgnoreCase("RULES", t.getType());
      boolean u = !StrUtil.equalsIgnoreCase("USER", t.getType());
      if (v && r && u) {
        if (StrUtil.startWithIgnoreCase(t.getType(), "ASSESSMENT_")) {
          formulaAsms.add(t);// 考核公式
        } else {
          formulaAccs.add(t);// 准确率合格率公式
        }
      }
      return v && r && u;
    }).collect(Collectors.toList());
    AssessmentRequest requestOfAsm = data.toBean(AssessmentRequest.class);// 考核计算参数
    CalculateRequest requestOfAcc = data.toBean(CalculateRequest.class);// 准确率计算参数
    requestOfAsm.setRatedCapacity(requestOfAcc.getElectricCapacity());
    for (CalculationFormula formulaAcc : formulaAccs) {
      Map<String, Object> accMap = new HashMap<>();
      CalculateResult accResult;
      formulaAcc.setIsRate("0");
      // 准确率、合格率
      requestOfAcc.setFormulaType(formulaAcc.getType());
      List<AssessmentRequest> powers = new ArrayList<>();
      if (StrUtil.containsAny(formulaAcc.getType(), "POINT_SHORT_ACCURACY")) { // 单点准确率
        List<CalculationInfo> points = requestOfAcc.getCalculationInfoList();
        List<CalculateResult> pointResults = new ArrayList<>();
        for (CalculationInfo point : points) { // 按点计算
          CalculateRequest pointRequest = BeanUtil.copyProperties(requestOfAcc, CalculateRequest.class);
          pointRequest.setCalculationInfoList(ListUtil.toList(point));
          CalculateResult pointResult = accuracyPassRateCalculateService.calculate(pointRequest, formulaAcc);
          pointResults.add(pointResult);
          AssessmentRequest asmRequest = new AssessmentRequest();
          boolean forecastEqualN99 = point.getForecastAbleValue().compareTo(new BigDecimal(-99)) == 0;
          boolean forecastEqualRei = point.getForecastAbleValue().compareTo(BigDecimal.ZERO) == 0;
          boolean realEqualN99 = point.getRealValue().compareTo(new BigDecimal(-99)) == 0;
          boolean realEqualRei = point.getRealValue().compareTo(BigDecimal.ZERO) == 0;
          asmRequest.setDeviationPower(!(forecastEqualN99 && forecastEqualRei) ? point.getForecastAbleValue() : !(realEqualN99 && realEqualRei) ? point.getRealValue() : BigDecimal.ZERO);
          powers.add(asmRequest);
        }
        accMap.put(formulaAcc.getType(), pointResults);
        result.add(accMap);
      } else { // 整日准确率
        // 调用计算方法
        accResult = accuracyPassRateCalculateService.calculate(requestOfAcc, formulaAcc);
        accMap.put(formulaAcc.getType(), ListUtil.toList(accResult));
        result.add(accMap);
      }
      if (StrUtil.containsAny(formulaAcc.getType(), "DAY_", "POINT_")) {
        String finalType = ReUtil.replaceAll(formulaAcc.getType(), "^DAY_|^POINT_", "");
        Optional<CalculationFormula> o = formulaAsms.stream().filter(t -> StrUtil.startWithIgnoreCase(StrUtil.removePrefix(t.getType(), "ASSESSMENT_"), finalType)).findFirst();
        CalculationFormula asmFormula = o.orElse(null);
        if (null != asmFormula) {
          List<CalculateResult> rateList = MapUtil.get(accMap, formulaAcc.getType(), List.class);
          List<CalculateResult> asmResults = new ArrayList<>();
          Map<String, Object> asmMap = new HashMap<>();
          for (int i = 0; i < rateList.size(); i++) {
            CalculateResult calculateResult = rateList.get(i);
            if (!powers.isEmpty()) {
              AssessmentRequest asmRequest = powers.get(i);
              requestOfAsm.setDeviationPower(asmRequest.getDeviationPower());
            }
            CalculateResult asmResult;
            if (NumberUtil.isNumber(ObjectUtil.toString(calculateResult.get("data")))) {
              BigDecimal rate = NumberUtil.toBigDecimal(ObjectUtil.toString(calculateResult.get("data")));
              if (StrUtil.startWith(finalType, "ULTRA_SHORT_ACCURACY")) {
                requestOfAsm.setUltraShortAccuracyRate(rate);
              } else if (StrUtil.startWith(finalType, "SHORT_ACCURACY")) {
                requestOfAsm.setShortAccuracyRate(rate);
              } else if (StrUtil.startWith(finalType, "MID_ACCURACY")) {
                requestOfAsm.setMidAccuracyRate(rate);
              }
            }
            requestOfAcc.setFormulaType(asmFormula.getType());
            asmResult = assessmentCalculateService.calculateAssessment(requestOfAsm);
            asmResults.add(asmResult);
          }
          asmMap.put(asmFormula.getType(), asmResults);
          result.add(asmMap);
        }
      }
    }
    System.out.println(JSONUtil.toJsonPrettyStr(result));
    return result;
  }

  /**
   * 计算单个公式准确率
   *
   * @param request
   * @return
   */
  @PostMapping("/compute")
  public ComputeReponse compute(HttpServletRequest request, @RequestBody JSONObject data) {
    // 校验信息
    ComputeReponse reponse = ComputeReponse.success("200", "Ok");
    reponse.clear();
    reponse.putAll(checkAuthorization(request));
    // 如果校验不成功，返回失败信息
    if (!ComputeReponse.Type.SUCCESS.value().equals(String.valueOf(reponse.get(CalculateResult.CODE_TAG)))) {
      return reponse;
    }
    ComputeRequest computeRequest = BeanUtil.toBean(data, ComputeRequest.class);
    ComputeResult computeResult = computeService.calculate(computeRequest);
    reponse.put("data", computeResult);
    System.out.println(JSONUtil.toJsonPrettyStr(reponse));
    // 调用计算方法
    return reponse;
  }

//    /**
//     * 计算所有公式准确率
//     *
//     * @param request
//     * @return
//     */
//    @PostMapping("/calculateAll")
//    public CalculateResult calculateAll(HttpServletRequest request, @RequestBody JSONObject data) {
//        // 校验信息
//        CalculateResult checkResult = checkAuthorization(request);
//        // 如果校验不成功，返回失败信息
//        if (!CalculateResult.Type.SUCCESS.value().equals(String.valueOf(checkResult.get(CalculateResult.CODE_TAG)))) {
//            return checkResult;
//        }
//        CalculateRequest calculateRequest;
//        // 将传入的数据转为bean
//        calculateRequest = data.toBean(CalculateRequest.class);
//        // 调用计算方法
//        CalculateResult resultMap = accuracyPassRateCalculateService.calculate(calculateRequest);
//        return resultMap;
//    }

  /**
   * 校验权限信息
   *
   * @param request
   * @return
   */
  public CalculateResult checkAuthorization(HttpServletRequest request) {
    String authorization = request.getHeader(CalculateResult.AUTHORIZATION);
    // 如果授权信息不存在，返回失败
    if (authorization == null) {
      return CalculateResult.error(CalculateResult.MSG_AUTHORIZATION_ERROR);
    }
    // 查询用户列表
    List<CalculationFormula> userList = calculationFormulaRepository.getUser();
    if (userList == null || userList.size() == 0) {
      return CalculateResult.error(CalculateResult.MSG_USERINFO_ERROR);
    }

    // 获取解密后的用户信息
    String deCodeUser = common.getBase64Decode(authorization);
    // 将数据库中的用户循环和传入的用户对比
    for (CalculationFormula calculationFormula : userList) {
      // 如果匹配上则返回成功，否则返回失败
      if (deCodeUser.equals(calculationFormula.getFormula())) {
        return CalculateResult.success();
      }
    }
    return CalculateResult.error(CalculateResult.MSG_AUTHORIZATION_ERROR);
  }

}
