package com.syjy.calculate.repository.repositoryImpl;

import cn.hutool.core.io.IoUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.db.sql.SqlUtil;
import cn.hutool.json.JSONUtil;
import cn.hutool.poi.excel.ExcelReader;
import cn.hutool.poi.excel.ExcelUtil;
import com.alibaba.druid.sql.SQLUtils;
import com.googlecode.aviator.AviatorEvaluator;
import com.syjy.calculate.config.StarterProperties;
import com.syjy.calculate.entity.CalculateResult;
import com.syjy.calculate.entity.CalculationFormula;
import com.syjy.calculate.entity.CalculationInfo;
import com.syjy.calculate.repository.CalculationFormulaRepository;
import com.syjy.calculate.util.DBUtil;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.jdbc.core.BatchPreparedStatementSetter;
import org.springframework.jdbc.core.BeanPropertyRowMapper;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;
import org.springframework.util.ClassUtils;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.io.InputStream;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * 公式计算
 *
 * @author zcl
 * @version 1.0
 * @since 2022/9/9 10:34
 */
@Service
@Slf4j
public class CalculationFormulaRepositoryImpl implements CalculationFormulaRepository {

  @Resource
  public JdbcTemplate jdbcTemplate;
  @Autowired
  StarterProperties properties;

  private String dbType;

  @PostConstruct
  public void init() {
    this.dbType = DBUtil.getDatabaseType(jdbcTemplate.getDataSource());
  }


  /**
   * 创建公式表
   */
  @Override
  public void initTable() {
    log.info("初始化准确率计算公式表");
    // 如果是金仓数据库，直接初始化表
    if (CalculateResult.DB_KINGBASE.equals(dbType)) {
      this.createTable();
      return;
    }

    // 从配置文件中获取版本
    String version = properties.getVersion();
    // 判断表是否存在
    String checkSql = "select count(1) from t_calculation_formula";
    int count = 0;
    try {
      // 获取表中数据
      count = jdbcTemplate.queryForObject(checkSql, Integer.class);
    } catch (DataAccessException e) {
      // 如果表不存在，则创建表
      log.error(e.toString());
      this.createTable();
    }

    // 再次查询表中数据
    try {
      count = jdbcTemplate.queryForObject(checkSql, Integer.class);
    } catch (DataAccessException e) {
      e.printStackTrace();
    }
    // 如果表存在
    if (count > 0) {
      // 从表中获取旧版本
      String oldVersion = getVersion();
      // 如果版本没变化，不进行初始化表操作
      if (oldVersion.equals(version) || oldVersion.equals("null")) {
        log.info("版本没变化，不进行初始化表");
        return;
      }
      // 如果版本不同，则初始化表
      this.createTable();
    }
  }

  @Override
  public int selectCount(String province, String type) {

    String sql = "SELECT count(1) from t_calculation_formula where state = '1' ";

    List<Object> param = new LinkedList<>();
    // 判断省调是否为空
    if (province != null && !"".equals(province.trim())) {
      sql = sql + "and province LIKE ? ";
      province = "%" + province + "%";
      param.add(province);
      if (type != null && !"".equals(type.trim())) {
        type = "%" + type + "%";
        param.add(type);
        sql = sql + "and type LIKE ? ";
      }
    } else {
      // 判断公式类型是否为空
      if (type != null && !"".equals(type.trim())) {
        type = "%" + type + "%";
        param.add(type);
        sql = sql + "and type LIKE ? ";
      }
    }
    int count = 0;
    try {
      if (param != null && param.size() > 0) {
        count = jdbcTemplate.queryForObject(sql, Integer.class, param.toArray());
      } else {
        count = jdbcTemplate.queryForObject(sql, Integer.class);
      }

    } catch (DataAccessException e) {
      e.printStackTrace();
    }
    return count;
  }

  /**
   * 查询所有公式
   *
   * @param limit    每页显示数量
   * @param offset   起始索引
   * @param province 省调
   * @param type     公式类型
   * @return
   */
  @Override
  public List<CalculationFormula> select(int limit, int offset, String province, String type) {

    String sql = "SELECT * from t_calculation_formula where state = '1' ";
    List<Object> param = new LinkedList<>();
    // 判断省调是否为空
    if (province != null && !"".equals(province.trim())) {
      sql = sql + "and province LIKE ? ";
      province = "%" + province + "%";
      param.add(province);
      if (type != null && !"".equals(type.trim())) {
        sql = sql + "and LIKE = ? ";
        type = "%" + type + "%";
        param.add(type);
      }
    } else {
      // 判断公式类型是否为空
      if (type != null && !"".equals(type.trim())) {
        sql = sql + "and type LIKE ? ";
        type = "%" + type + "%";
        param.add(type);
      }
    }
    // 拼接分页范围
    sql = sql + " limit " + offset + "," + limit;
    List<CalculationFormula> calculationFormulaList = new ArrayList<>();

    try {
      if (param != null && param.size() > 0) {
        calculationFormulaList = jdbcTemplate.query(sql, new BeanPropertyRowMapper<>(CalculationFormula.class), param.toArray());
      } else {
        calculationFormulaList = jdbcTemplate.query(sql, new BeanPropertyRowMapper<>(CalculationFormula.class));
      }
    } catch (DataAccessException e) {
      e.printStackTrace();
    }
    return calculationFormulaList;
  }

  @Override
  public CalculationFormula getOne(int id) {
    String sql = "select * from t_calculation_formula where id = ? and state = '1' ";
    CalculationFormula calculationFormula;
    try {
      calculationFormula = jdbcTemplate.queryForObject(sql, new BeanPropertyRowMapper<>(CalculationFormula.class), id);
    } catch (EmptyResultDataAccessException e) {
      return null;
    }
    return calculationFormula;
  }

  /**
   * 更新公式
   *
   * @param calculationFormula
   * @return
   */
  @Override
  public int update(CalculationFormula calculationFormula) {
    Integer id = calculationFormula.getId();
    String type = calculationFormula.getType();
    String formula = calculationFormula.getFormula();
    String ruleFormula = calculationFormula.getRuleFormula();
    String province = calculationFormula.getProvince();
    String electricType = calculationFormula.getElectricType();
    String isRate = calculationFormula.getIsRate();
    int count = 0;
    try {
      count = jdbcTemplate.update("update t_calculation_formula set TYPE=?,FORMULA=?,PROVINCE=?,ELECTRIC_TYPE=?,RULE_FORMULA = ?,IS_RATE=? where id=?", type, formula, province, electricType, ruleFormula, isRate, id);
    } catch (DataAccessException e) {
      e.printStackTrace();
    }
    if (count > 0) {
      // 清除缓存公式
      AviatorEvaluator.clearExpressionCache();
    }
    return count;
  }

  /**
   * 新增公式
   *
   * @param calculationFormula
   * @return
   */
  @Override
  public int add(CalculationFormula calculationFormula) {
    String electricType = calculationFormula.getElectricType();
    String isRate = calculationFormula.getIsRate();
    String type = calculationFormula.getType();
    String formula = calculationFormula.getFormula();
    String ruleFormula = calculationFormula.getRuleFormula();
    String province = calculationFormula.getProvince();
    int count = 0;
    try {
      count = jdbcTemplate.update(
        "INSERT INTO t_calculation_formula(TYPE,FORMULA,PROVINCE,ELECTRIC_TYPE,RULE_FORMULA,IS_RATE,STATE) VALUES (?, ?, ?, ?, ?, ?,'1')", type, formula, province, electricType, ruleFormula, isRate);
    } catch (DataAccessException e) {
      e.printStackTrace();
    }
    return count;
  }

  /**
   * 删除公式
   *
   * @param id
   * @return
   */
  @Override
  public int delete(int id) {
    int count = 0;
    try {
      count = jdbcTemplate.update("update t_calculation_formula set state = 0 where id =?", id);
    } catch (DataAccessException e) {
      e.printStackTrace();
    }
    return count;
  }

  /**
   * 查询公式数据
   *
   * @param type         类型
   * @param province     省调
   * @param formulaType  公式类型
   * @param electricType 场站类型
   * @return
   */
  @Override
  public CalculationFormula findByTypeAndProvince(String type, String formulaType, String province, String electricType, String stationCode) {
    String sql = "";
    electricType = "%" + electricType + "%";
    if (stationCode != null && !"".equals(stationCode)) {
      stationCode = "%" + stationCode + "%";
    }
    // 查询sql
    if (CalculateResult.FORMULA.equals(type)) {
      sql = "SELECT * from t_calculation_formula where state = '1' and ELECTRIC_TYPE like ? and PROVINCE = ? and TYPE = ? and STATION_CODE like ? limit 1 ";
    } else if (CalculateResult.RULES.equals(type)) {
      sql = "SELECT * from t_calculation_formula where state = '1' and TYPE = 'RULES' and ELECTRIC_TYPE like ? and PROVINCE = ? and RULE_FORMULA = ? and STATION_CODE like ? limit 1";
    } else {
      log.error(CalculateResult.MSG_FORMULA_TYPE_ERROR);
    }

    CalculationFormula calculationFormula = new CalculationFormula();
    // 根据类型和省调查询公式
    try {
      calculationFormula = jdbcTemplate.queryForObject(sql, new BeanPropertyRowMapper<>(CalculationFormula.class), electricType, province, formulaType, stationCode);
      return calculationFormula;
    } catch (EmptyResultDataAccessException e) {
      // 如果根据场站编码没查到，则在根据省调和场站类型是All的查
      if (CalculateResult.FORMULA.equals(type)) {
        sql = "SELECT * from t_calculation_formula where state = '1' and ELECTRIC_TYPE like ? and PROVINCE = ? and TYPE = ?  and STATION_CODE = 'ALL' limit 1";
      } else if (CalculateResult.RULES.equals(type)) {
        sql = "SELECT * from t_calculation_formula where state = '1' and TYPE = 'RULES' and ELECTRIC_TYPE like ? and PROVINCE = ? and RULE_FORMULA = ? and STATION_CODE = 'ALL' limit 1";
      }
      // 根据类型和省调查询公式
      try {
        calculationFormula = jdbcTemplate.queryForObject(sql, new BeanPropertyRowMapper<>(CalculationFormula.class), electricType, province, formulaType);
        return calculationFormula;
      } catch (EmptyResultDataAccessException ere) {
        // 如果根据场站编码没查到，则在根据省调和场站类型是All的查
        return null;
      }
    }
  }

  @Override
  public List<CalculationFormula> findFormulaByProvince(String province, String electricType) {
    List<CalculationFormula> result = new ArrayList<>();
    electricType = "%" + electricType + "%";
    // 查询sql
    String sql = " SELECT * from t_calculation_formula where state = '1' and ELECTRIC_TYPE like ? and PROVINCE = ? and RULE_FORMULA is null and INSTR(`TYPE` ,'ASSESSMENT_') <> 1 ";
    try {
      result = jdbcTemplate.query(sql, new BeanPropertyRowMapper<>(CalculationFormula.class), electricType, province);
    } catch (Exception ex) {
      ex.printStackTrace();
      log.error(CalculateResult.MSG_NO_FORMULA);
    }
    return result;
  }

  public List<CalculationFormula> findFormulaWithProvince(String province, String electricType) {
    List<CalculationFormula> result = new ArrayList<>();
    electricType = "%" + electricType + "%";
    // 查询sql
    String sql = " SELECT * from t_calculation_formula where state = '1' and ELECTRIC_TYPE like ? and PROVINCE = ? ";
    try {
      result = jdbcTemplate.query(sql, new BeanPropertyRowMapper<>(CalculationFormula.class), electricType, province);
    } catch (Exception ex) {
      ex.printStackTrace();
      log.error(CalculateResult.MSG_NO_FORMULA);
    }
    return result;
  }

  /**
   * 查询考核公式数据
   *
   * @param type         类型
   * @param province     省调
   * @param formulaType  公式类型
   * @param electricType 场站类型
   * @return
   */
  @Override
  public CalculationFormula findAssessmentByTypeAndProvince(String type, String formulaType, String province, String electricType, String stationCode) {

    electricType = "%" + electricType + "%";
    if (stationCode != null && !"".equals(stationCode)) {
      stationCode = "%" + stationCode + "%";
    }
    String sql = "";

    // 查询sql
    if (CalculateResult.FORMULA.equals(type)) {
      sql = "SELECT * from t_calculation_formula where state = '1' and ELECTRIC_TYPE like ? and PROVINCE = ? and TYPE = ?  and STATION_CODE like ? limit 1";
    } else if (CalculateResult.RULES.equals(type)) {
      sql = "SELECT * from t_calculation_formula where state = '1' and TYPE = 'RULES' and ELECTRIC_TYPE like ? and PROVINCE = ? and RULE_FORMULA = ? and STATION_CODE like ? limit 1";
    } else {
      log.error(CalculateResult.MSG_FORMULA_TYPE_ERROR);
      return null;
    }
    // 根据类型和省调查询公式
    try {
      CalculationFormula calculationFormula = jdbcTemplate.queryForObject(sql, new BeanPropertyRowMapper<>(CalculationFormula.class), electricType, province, formulaType, stationCode);
      return calculationFormula;
    } catch (EmptyResultDataAccessException e) {
      // 如果根据场站编码没查到，则在根据省调和场站类型是All的查
      // 查询sql
      if (CalculateResult.FORMULA.equals(type)) {
        sql = "SELECT * from t_calculation_formula where state = '1' and ELECTRIC_TYPE like ? and PROVINCE = ? and TYPE = ?  and STATION_CODE = 'ALL' limit 1";
      } else if (CalculateResult.RULES.equals(type)) {
        sql = "SELECT * from t_calculation_formula where state = '1' and TYPE = 'RULES' and ELECTRIC_TYPE like ? and PROVINCE = ? and RULE_FORMULA = ? and STATION_CODE = 'ALL' limit 1";
      }
      // 根据类型和省调查询公式
      try {
        CalculationFormula calculationFormula = jdbcTemplate.queryForObject(sql, new BeanPropertyRowMapper<>(CalculationFormula.class), electricType, province, formulaType);
        return calculationFormula;
      } catch (EmptyResultDataAccessException ere) {
        // 如果根据场站编码没查到，则在根据省调和场站类型是All的查
        return null;
      }
    }
  }

  /**
   * 获取测试数据
   */
  @Override
  public List<CalculationInfo> getCalculationInfo() {
    String sql = "SELECT * FROM `t_calculation_info`";
    List<CalculationInfo> list = new ArrayList<>();
    try {
      list = jdbcTemplate.query(sql, new BeanPropertyRowMapper<>(CalculationInfo.class));
      return list;
    } catch (EmptyResultDataAccessException ere) {
      return null;
    }
  }

  /**
   * 创建表
   */
  private void createTable() {
    InputStream excelInputStream = null;
    InputStream sqlInputStream = null;
    try {
      log.info("---------------开始创建准确率计算表---------------");
      // 从配置文件中获取版本

      String path;
      // 获取初始化公式表sql
      switch (dbType) {
        case CalculateResult.DB_KINGBASE: // 金昌数据库
          path = "sql/t_calculation_formula_king_base.sql";
          break;
        case CalculateResult.DB_POSTGRES: // PG数据库
          path = "sql/t_calculation_formula_postgresql.sql";
          break;
        case CalculateResult.DB_HIGHGO: // 瀚高数据库
          path = "sql/t_calculation_formula_highgo.sql";
          break;
        case CalculateResult.DB_XUGU: // 虚谷数据库
          path = "sql/t_calculation_formula_xugu.sql";
          break;
        case CalculateResult.DB_MYSQL: // Mysql数据库
        default:
          path = "sql/t_calculation_formula.sql";
      }
      log.info("建表：{}", path);
      // 用ClassLoader获取jar包中的文件，获取文件流
      ClassLoader defaultClassLoader = ClassUtils.getDefaultClassLoader();
      assert defaultClassLoader != null;
      URL resource = defaultClassLoader.getResource(path);
      assert resource != null;
      sqlInputStream = resource.openStream();
      //
      URL excelUrl = defaultClassLoader.getResource("sql/t_calculation_formula.xlsx");
      if (null == excelUrl) {
        return;
      }
      excelInputStream = excelUrl.openStream();
      ExcelReader excelReader = ExcelUtil.getReader(excelInputStream, 0, true);
      List<CalculationFormula> formulas = excelReader.readAll(CalculationFormula.class);
      BatchPreparedStatementSetter pss = new BatchPreparedStatementSetter() {
        @Override
        public void setValues(PreparedStatement ps, int i) throws SQLException {
          CalculationFormula formula = formulas.get(i);
          int id = i + 1;
          ps.setInt(1, id);
          ps.setString(2, formula.getType());
          ps.setInt(3, ObjectUtil.defaultIfNull(formula.getOrderNo(), 1));
          //String formula = StrUtil.removeAll(formula.getFormula(), '\r', '\n');
          ps.setObject(4, formula.getFormula());
          ps.setString(5, formula.getProvince());
          ps.setString(6, formula.getElectricType());
          ps.setString(7, formula.getStationCode());
          ps.setString(8, formula.getRuleFormula());
          ps.setString(9, formula.getIsRate());
          ps.setString(10, formula.getMaxMinLimit());
          ps.setString(11, formula.getState());
          ps.setString(12, formula.getDataSourceType());
          ps.setString(13, formula.getHowManyDays());
          ps.setObject(14, formula.getCreateTime());
          ps.setString(15, formula.getCreator());
          ps.setString(16, formula.getLastModifier());
          ps.setObject(17, formula.getLastModifyTime());
          log.info("PS ## {} ## {}", id, ps);
        }

        @Override
        public int getBatchSize() {
          return formulas.size();
        }
      };
      String sql = "INSERT INTO T_CALCULATION_FORMULA (ID, `TYPE`, ORDER_NO, FORMULA, PROVINCE, ELECTRIC_TYPE, STATION_CODE, RULE_FORMULA, IS_RATE, MAX_MIN_LIMIT, STATE, DATA_SOURCE_TYPE, HOW_MANY_DAYS, CREATE_TIME, CREATOR, LAST_MODIFIER, LAST_MODIFY_TIME) VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
      switch (dbType) {
        case CalculateResult.DB_KINGBASE:
        case CalculateResult.DB_POSTGRES:
        case CalculateResult.DB_HIGHGO:
        case CalculateResult.DB_XUGU:
          String initSQL = SqlUtil.formatSql(IoUtil.read(sqlInputStream, StandardCharsets.UTF_8));
          List<String> ddls = StrUtil.splitTrim(initSQL, ";");
          for (String ddl : ddls) {
            if (!ddl.trim().isEmpty()) {
              jdbcTemplate.execute(ddl);
            }
          }
          jdbcTemplate.batchUpdate(SQLUtils.format(sql, dbType), pss);
          break;
        case CalculateResult.DB_MYSQL:
          String ddl = "SET NAMES utf8mb4;\nSET FOREIGN_KEY_CHECKS = 0;\n" + IoUtil.read(sqlInputStream, StandardCharsets.UTF_8);
          jdbcTemplate.execute(SQLUtils.formatMySql(ddl));
          jdbcTemplate.batchUpdate(SQLUtils.format(sql, dbType), pss);
          jdbcTemplate.execute(SQLUtils.formatMySql("SET FOREIGN_KEY_CHECKS = 1;"));
          break;
        default:
          throw new Exception("未知数据库类型");
      }
      // 执行sql文件
      log.info("建表成功");
      //
    } catch (Exception ex) {
      log.error("建表失败：{} ", ex);
    } finally {
      IoUtil.closeIfPosible(excelInputStream);
      IoUtil.closeIfPosible(sqlInputStream);
    }
    log.info("---------------结束创建准确率计算表---------------");
    // 更新公式
    updateVersion();
  }

  /**
   * 获取版本
   */
  private String getVersion() {
    String version = "";

    String sql = " SELECT * FROM t_calculation_formula where type = 'VERSION' ";
    switch (dbType) {
      case CalculateResult.DB_KINGBASE:
      case CalculateResult.DB_POSTGRES:
      case CalculateResult.DB_HIGHGO:
      case CalculateResult.DB_XUGU:
        sql = " SELECT * FROM t_calculation_formula where \"TYPE\" = 'VERSION' ";
        break;
      default:
        sql = " SELECT * FROM t_calculation_formula where type = 'VERSION' ";
    }
    try {
      CalculationFormula calculationFormula = jdbcTemplate.queryForObject(sql, new BeanPropertyRowMapper<>(CalculationFormula.class));
      version = calculationFormula.getFormula();
    } catch (DataAccessException e) {
      e.printStackTrace();
    }
    return version;
  }

  /**
   * 获取用户密码
   */
  @Override
  public List<CalculationFormula> getUser() {

    String sql = " SELECT * FROM t_calculation_formula where type = 'USER' ";
    switch (dbType) {
      case CalculateResult.DB_KINGBASE:
      case CalculateResult.DB_POSTGRES:
      case CalculateResult.DB_HIGHGO:
      case CalculateResult.DB_XUGU:
        sql = " SELECT * FROM t_calculation_formula where \"TYPE\" = 'USER' ";
        break;
      default:
        sql = " SELECT * FROM t_calculation_formula where type = 'USER' ";
    }
    List<CalculationFormula> list = new ArrayList<>();
    try {
      list = jdbcTemplate.query(sql, new BeanPropertyRowMapper<>(CalculationFormula.class));
      return list;
    } catch (EmptyResultDataAccessException ere) {
      return null;
    }
  }

  /**
   * 更新版本
   */
  private void updateVersion() {
    String version = properties.getVersion();

    String sql = " UPDATE t_calculation_formula set FORMULA = '" + version + "' where type = 'VERSION' ";
    switch (dbType) {
      case CalculateResult.DB_KINGBASE:
      case CalculateResult.DB_POSTGRES:
      case CalculateResult.DB_HIGHGO:
      case CalculateResult.DB_XUGU:
        sql = " UPDATE t_calculation_formula SET \"FORMULA\"='" + version + "' WHERE \"TYPE\"='VERSION' ";
        break;
      default:
        sql = " UPDATE t_calculation_formula set FORMULA = '" + version + "' where type = 'VERSION' ";
    }
    try {
      jdbcTemplate.update(sql);
    } catch (DataAccessException e) {
      e.printStackTrace();
    }
  }

}
